"""Config file loading and parser defaults application for CLIs."""
# pylint: disable=too-many-return-statements

import argparse
import fnmatch
import json
from argparse import Namespace
from pathlib import Path
from typing import Any, Dict, Optional

import yaml

from verbatim_files.format.writer import LanguageStyle, ProbabilityStyle, SpeakerStyle, TimestampStyle

OUTPUT_FLAGS = ("ass", "docx", "txt", "json", "jsonl", "md", "stdout", "stdout_nocolor")
DEFAULT_MATCH = ["*.wav", "*.mp3", "*.m4a", "*.mp4"]


def load_config_file(path: str) -> Dict[str, Any]:
    config_path = Path(path).expanduser()
    if not config_path.exists():
        raise FileNotFoundError(f"Config file not found: {config_path}")

    suffix = config_path.suffix.lower()
    if suffix in (".yml", ".yaml"):
        with config_path.open("r", encoding="utf-8") as f:
            return yaml.safe_load(f) or {}
    if suffix == ".json":
        with config_path.open("r", encoding="utf-8") as f:
            return json.load(f)
    raise ValueError(f"Unsupported config file format: {suffix}")


def _normalize_styles(cfg: Dict[str, Any]) -> Dict[str, Any]:
    normalized = dict(cfg)
    if "format_timestamp" in normalized and isinstance(normalized["format_timestamp"], str):
        normalized["format_timestamp"] = TimestampStyle[normalized["format_timestamp"]]
    if "format_speaker" in normalized and isinstance(normalized["format_speaker"], str):
        normalized["format_speaker"] = SpeakerStyle[normalized["format_speaker"]]
    if "format_probability" in normalized and isinstance(normalized["format_probability"], str):
        normalized["format_probability"] = ProbabilityStyle[normalized["format_probability"]]
    if "format_language" in normalized and isinstance(normalized["format_language"], str):
        normalized["format_language"] = LanguageStyle[normalized["format_language"]]
    return normalized


def _flatten_output(output_section: Dict[str, Any]) -> Dict[str, Any]:
    flat: Dict[str, Any] = {}
    formats = None
    if "formats" in output_section and isinstance(output_section["formats"], list):
        formats = output_section["formats"]
    if "files" in output_section and isinstance(output_section["files"], list):
        formats = output_section["files"]
    if formats is not None:
        selected = set(formats)
        for flag in OUTPUT_FLAGS:
            flat[flag] = flag.replace("_", "-") in selected or flag in selected
    else:
        for flag in OUTPUT_FLAGS:
            if flag in output_section:
                flat[flag] = output_section[flag]
    return flat


def _flatten_format(format_section: Dict[str, Any]) -> Dict[str, Any]:
    flat: Dict[str, Any] = {}
    for key in ("timestamp", "speaker", "probability", "language"):
        fmt_key = f"format_{key}"
        if key in format_section:
            flat[fmt_key] = format_section[key]
    return flat


def flatten_profile(profile: Dict[str, Any]) -> Dict[str, Any]:
    flat: Dict[str, Any] = {}
    for key, val in profile.items():
        if key == "match":
            if isinstance(val, dict):
                filenames = val.get("filename")
                if filenames:
                    flat["match"] = filenames
            elif isinstance(val, list):
                flat["match"] = val
            continue
        if key == "include" and isinstance(val, dict):
            filenames = val.get("filename")
            if filenames:
                flat["match"] = filenames
            continue
        if key == "ignore" and isinstance(val, dict):
            filenames = val.get("filename")
            if filenames:
                flat["ignore"] = filenames
            continue
        if key == "output" and isinstance(val, dict):
            flat.update(_flatten_output(val))
            continue
        if key == "format" and isinstance(val, dict):
            flat.update(_flatten_format(val))
            continue
        flat[key] = val
    return _normalize_styles(flat)


def _match_profile(profile: Dict[str, Any], filename: Optional[str]) -> bool:
    includes = []
    excludes = []
    if isinstance(profile.get("include"), dict):
        includes = profile["include"].get("filename") or []
    if isinstance(profile.get("ignore"), dict):
        excludes = profile["ignore"].get("filename") or []

    if filename is None:
        return True
    name = Path(filename).name
    if includes:
        if not any(fnmatch.fnmatch(name, pat) for pat in includes):
            return False
    if excludes and any(fnmatch.fnmatch(name, pat) for pat in excludes):
        return False
    return True


def select_profile(config_data: Dict[str, Any], filename: Optional[str]) -> Dict[str, Any]:
    if not config_data:
        return {}

    base_defaults = flatten_profile({k: v for k, v in config_data.items() if k != "profiles"})
    if "match" not in base_defaults:
        base_defaults["match"] = DEFAULT_MATCH
    profiles = config_data.get("profiles")
    if filename is None and profiles:
        # For global defaults (no filename), ignore profile matching and return root defaults only.
        return base_defaults
    if isinstance(profiles, list) and profiles:
        fallback: Optional[Dict[str, Any]] = None
        for profile in profiles:
            if not isinstance(profile, dict):
                continue
            if _match_profile(profile, filename):
                merged = base_defaults.copy()
                merged.update(flatten_profile(profile))
                return merged
            if profile.get("match") is None and fallback is None:
                fallback = profile
        if fallback:
            merged = base_defaults.copy()
            merged.update(flatten_profile(fallback))
            return merged
        return base_defaults

    # Single-profile config; honor match at root
    if filename is not None and not _match_profile(config_data, filename):
        return {}
    merged = base_defaults.copy()
    merged.update(flatten_profile(config_data))
    return merged


def merge_args(base_defaults: Namespace, profile_overrides: Dict[str, Any], user_args: Namespace) -> Namespace:
    merged = vars(base_defaults).copy()
    merged.update(profile_overrides)
    for key, val in vars(user_args).items():
        # Only override when the user provided a value different from the parser default
        if key in merged and val == getattr(base_defaults, key, None):
            continue
        merged[key] = val
    # If match/ignore not specified by user or profile, ensure defaults
    if "match" not in merged or merged["match"] is None:
        merged["match"] = DEFAULT_MATCH
    if "ignore" not in merged or merged["ignore"] is None:
        merged["ignore"] = []
    return argparse.Namespace(**merged)



import logging
import warnings
from abc import ABC, abstractmethod
from typing import Dict, List

import torch
from numpy.typing import NDArray
from silero_vad import get_speech_timestamps, load_silero_vad

# Configure logger
LOG = logging.getLogger(__name__)


class VoiceActivityDetection(ABC):
    @abstractmethod
    def find_activity(
        self,
        audio: NDArray,
        min_speech_duration_ms: int = 250,
        min_silence_duration_ms: int = 100,
    ) -> List[Dict[str, int]]:
        pass


class SileroVoiceActivityDetection(VoiceActivityDetection):
    def __init__(self):
        with warnings.catch_warnings():
            # suppresses silero_vad\model.py:15: DeprecationWarning: path is deprecated. Use files() instead.
            # Refer to https://importlib-resources.readthedocs.io/en/latest/using.html#migrating-from-legacy for migration advice.
            warnings.filterwarnings("ignore", category=DeprecationWarning)
            self.model = load_silero_vad()

    def find_activity(
        self,
        audio: NDArray,
        min_speech_duration_ms: int = 250,
        min_silence_duration_ms: int = 100,
    ) -> List[Dict[str, int]]:
        audio_tensor = torch.from_numpy(audio).float()
        speech_timestamps = get_speech_timestamps(
            audio=audio_tensor,
            model=self.model,
            threshold=0.25,
            sampling_rate=16000,
            min_speech_duration_ms=min_speech_duration_ms,
            max_speech_duration_s=float("inf"),
            min_silence_duration_ms=min_silence_duration_ms,
            speech_pad_ms=250,
            return_seconds=False,
            visualize_probs=False,
            progress_tracking_callback=None,  # pyright: ignore[reportArgumentType]
            neg_threshold=None,  # pyright: ignore[reportArgumentType]
            window_size_samples=512,
        )
        return speech_timestamps



import json
import logging
import re
from pathlib import Path
from typing import Dict, List, Tuple

from openai import OpenAI

from verbatim.transcript.postprocessing.config import Config

LOG = logging.getLogger(__name__)


class DiarizationProcessor:
    def __init__(self, config: Config):
        """Initialize the processor with configuration"""
        self.config = config
        self.client = OpenAI(base_url=config.OLLAMA_BASE_URL, api_key=config.API_KEY)

    def extract_text_and_spk(self, completion: str) -> Tuple[str, str]:
        """Extract text and speaker labels from completion string"""
        spk = "1"
        previous_spk = "1"
        result_text = []
        result_spk = []

        for word in completion.split():
            if word.startswith("<speaker:"):
                if not word.endswith(">"):
                    word += ">"
                spk = word[len("<speaker:") : -len(">")]
                try:
                    spk_int = int(spk)
                    if not spk or spk_int < 1 or spk_int > 10:
                        raise ValueError(f"Unexpected speaker token: {word}")
                    previous_spk = spk
                except ValueError:
                    LOG.warning(f"Skipping meaningless speaker token: {word}")
                    spk = previous_spk
            else:
                result_text.append(word)
                result_spk.append(spk)

        return " ".join(result_text), " ".join(result_spk)

    def process_chunk(self, text: str) -> Tuple[str, str]:
        """Process a single chunk of text"""
        try:
            response = self.client.chat.completions.create(
                model=self.config.MODEL_NAME,
                messages=[{"role": "system", "content": self.config.SYSTEM_PROMPT}, {"role": "user", "content": f"{text} -->"}],
                temperature=0.1,
            )

            completion = response.choices[0].message.content or ""

            # Create and log pretty diff
            print("\nProcessing chunk:")
            print("=" * 80)
            print("Before:")
            print(text)  # Original formatted text
            print("\nAfter:")
            print(completion)

            return self.extract_text_and_spk(completion)

        except Exception as e:
            LOG.error(f"Error processing chunk: {e}")
            raise

    def clean_speaker_tag(self, tag: str) -> str:
        """Clean up speaker tags by removing repeated numbers"""
        # Extract the first number from the tag
        match = re.search(r"<speaker:(\d+)", tag)
        if match:
            number = match.group(1)
            return f"<speaker:{number}>"
        return tag

    def format_chunk(self, utterances: List[Dict]) -> str:
        """Format a chunk of utterances into diarized text"""
        # Join utterances with proper speaker tags
        text_parts = []
        for utt in utterances:
            # Clean up any repeated speaker numbers in the input
            speaker_tag = self.clean_speaker_tag(f"<speaker:{utt['hyp_spk']}>")
            text_parts.append(f"{speaker_tag} {utt['hyp_text']}")

        return "\n".join(text_parts)

    def process_json(self, input_path: Path, output_path: Path, chunk_size: int = 3) -> Dict:
        """Process entire JSON file and save results"""
        with open(input_path) as f:
            data = json.load(f)

        output_utterances = []
        current_chunk = []

        for utterance in data["utterances"]:
            current_chunk.append(utterance)

            if len(current_chunk) >= chunk_size:
                chunk_text = self.format_chunk(current_chunk)
                text, spk = self.process_chunk(chunk_text)

                output_utterances.append({"utterance_id": f"utt{len(output_utterances)}", "hyp_text": text, "hyp_spk": spk})
                current_chunk = []

        # Process remaining utterances
        if current_chunk:
            chunk_text = self.format_chunk(current_chunk)
            text, spk = self.process_chunk(chunk_text)
            output_utterances.append({"utterance_id": f"utt{len(output_utterances)}", "hyp_text": text, "hyp_spk": spk})

        output_data = {"utterances": output_utterances}

        with open(output_path, "w", encoding="utf-8") as f:
            json.dump(output_data, f, ensure_ascii=False, indent=2)

        return output_data



"""Single-file execution helpers for the CLI."""
# pylint: disable=import-outside-toplevel

from typing import List

from verbatim.verbatim import execute
from verbatim_audio.sources.audiosource import AudioSource


def build_audio_sources(
    *,
    args,
    config,
    source_config,
    source_path: str,
    working_prefix_no_ext: str,
    output_prefix_no_ext: str,
) -> List[AudioSource]:
    from verbatim_audio.sources.factory import create_audio_sources

    return create_audio_sources(
        source_config=source_config,
        device=config.device,
        cache=config.cache,
        input_source=source_path,
        start_time=args.start_time,
        stop_time=args.stop_time,
        working_prefix_no_ext=working_prefix_no_ext,
        output_prefix_no_ext=output_prefix_no_ext,
        stream=config.stream,
    )


def run_execute(
    *,
    source_path: str,
    config,
    write_config,
    audio_sources: List[AudioSource],
    output_formats,
    output_prefix_no_ext: str,
    working_prefix_no_ext: str,
    eval_file,
):
    execute(
        source_path=source_path,
        config=config,
        write_config=write_config,
        audio_sources=audio_sources,
        output_formats=output_formats,
        output_prefix_no_ext=output_prefix_no_ext,
        working_prefix_no_ext=working_prefix_no_ext,
        eval_file=eval_file,
    )



import logging
from abc import ABC, abstractmethod
from typing import List, Optional, Tuple

from numpy.typing import NDArray

from ...transcript.words import Word

LOG = logging.getLogger(__name__)

PREPEND_PUNCTUATIONS = "\"'“¿([{-"
APPEND_PUNCTUATIONS = "\"'.。,;，!！?？:：”)]}、"


class WhisperConfig:
    def __init__(self):
        self.prepend_punctuations = PREPEND_PUNCTUATIONS
        self.append_punctuations = APPEND_PUNCTUATIONS
        self.task = "transcribe"
        self.length_penalty = 1.0
        self.suppress_tokens = [-1]
        self.suppress_blank = False
        self.no_speech_threshold = 0.95
        self.logprob_threshold = -1.0
        self.compression_ratio_threshold = 2.4


class Transcriber(ABC):
    @abstractmethod
    def guess_language(self, audio: NDArray, lang: List[str]) -> Tuple[str, float]:
        pass

    @abstractmethod
    def transcribe(
        self,
        *,
        audio: NDArray,
        lang: str,
        prompt: str,
        prefix: str,
        window_ts: int,
        audio_ts: int,
        whisper_beam_size: int = 3,
        whisper_best_of: int = 3,
        whisper_patience: float = 1.0,
        whisper_temperatures: Optional[List[float]] = None,
    ) -> List[Word]:
        pass



from dataclasses import dataclass
from typing import List, Optional, TextIO

import numpy as np
from colorama import Fore, Style

from verbatim.transcript.words import Utterance, Word
from verbatim_diarization import UNKNOWN_SPEAKER

from ..formatting import format_milliseconds
from .writer import (
    LanguageStyle,
    ProbabilityStyle,
    SpeakerStyle,
    TimestampStyle,
    TranscriptWriter,
    TranscriptWriterConfig,
)


@dataclass
class ColorScheme:
    color_timestamp: str = Fore.CYAN
    color_speaker: str = Fore.BLUE
    color_language: str = Fore.YELLOW
    color_text: str = Fore.GREEN
    color_text_lowconfidence: str = Fore.LIGHTYELLOW_EX
    color_text_verylowconfidence: str = Fore.LIGHTRED_EX
    color_reset: str = Style.RESET_ALL


COLORSCHEME_ACKNOWLEDGED = ColorScheme(
    color_timestamp=Fore.LIGHTCYAN_EX,
    color_speaker=Fore.LIGHTBLUE_EX,
    color_language=Fore.LIGHTYELLOW_EX,
    color_text=Fore.LIGHTGREEN_EX,
    color_text_lowconfidence=Fore.LIGHTYELLOW_EX,
    color_text_verylowconfidence=Fore.LIGHTRED_EX,
    color_reset=Style.RESET_ALL,
)

COLORSCHEME_UNACKNOWLEDGED = ColorScheme(
    color_timestamp=Fore.CYAN,
    color_speaker=Fore.BLUE,
    color_language=Fore.YELLOW,
    color_text=Fore.GREEN,
    color_text_lowconfidence=Fore.LIGHTYELLOW_EX,
    color_text_verylowconfidence=Fore.LIGHTRED_EX,
    color_reset=Style.RESET_ALL,
)

COLORSCHEME_UNCONFIRMED = ColorScheme(
    color_timestamp=Fore.CYAN,
    color_speaker=Fore.BLUE,
    color_language=Fore.YELLOW,
    color_text=Fore.LIGHTBLACK_EX,
    color_text_lowconfidence=Fore.LIGHTBLACK_EX,
    color_text_verylowconfidence=Fore.LIGHTBLACK_EX,
    color_reset=Style.RESET_ALL,
)

COLORSCHEME_NONE = ColorScheme(
    color_timestamp="",
    color_speaker="",
    color_language="",
    color_text="",
    color_reset="",
    color_text_lowconfidence="",
    color_text_verylowconfidence="",
)


class TranscriptFormatter:
    def __init__(
        self,
        speaker_style: SpeakerStyle,
        timestamp_style: TimestampStyle,
        probability_style: ProbabilityStyle,
        language_style: LanguageStyle,
    ):
        self.current_language = None
        self.current_speaker = None
        self.speaker_style = speaker_style
        self.timestamp_style = timestamp_style
        self.probability_style = probability_style
        self.language_style = language_style
        self.last_ts: int = 0

    def _format_timestamp(self, out: TextIO, start_ts: int, end_ts: int, colours: ColorScheme):
        if self.timestamp_style == TimestampStyle.none:
            pass

        elif self.timestamp_style == TimestampStyle.minute:
            if self.last_ts is None or start_ts - self.last_ts >= 60 * 16000:
                self.last_ts = ((start_ts // 16000) // 60) * 60 * 16000
                out.write("\n")
                out.write(colours.color_timestamp)
                out.write(f"[{format_milliseconds(start_ts * 1000 / 16000)}]")
                out.write(colours.color_reset)
                out.write("\n\n")

        elif self.timestamp_style == TimestampStyle.start:
            out.write(colours.color_timestamp)
            out.write(f"[{format_milliseconds(start_ts * 1000 / 16000)}]:")
            out.write(colours.color_reset)

        elif self.timestamp_style == TimestampStyle.range:
            out.write(colours.color_timestamp)
            out.write(f"[{format_milliseconds(start_ts * 1000 / 16000)}-{format_milliseconds(end_ts * 1000 / 16000)}]:")
            out.write(colours.color_reset)

    def _format_speaker(self, out: TextIO, speaker: str, colours: ColorScheme):
        if self.speaker_style == SpeakerStyle.none:
            return

        elif self.speaker_style == SpeakerStyle.change:
            if speaker != self.current_speaker:
                self.current_speaker = speaker
                out.write(colours.color_speaker)
                out.write(f"[{speaker}]")
                out.write(colours.color_reset)

        elif self.speaker_style == SpeakerStyle.always:
            out.write(colours.color_speaker)
            out.write(f"[{speaker}]")
            out.write(colours.color_reset)

    def _format_language(self, out: TextIO, language: str, first_word: bool, colours: ColorScheme):
        if self.language_style == LanguageStyle.none:
            pass
        elif self.language_style == LanguageStyle.change:
            if language != self.current_language:
                self.current_language = language
                out.write(colours.color_language)
                out.write(f"[{language}]")
                out.write(colours.color_reset)
        elif self.language_style == LanguageStyle.always:
            if first_word or language != self.current_language:
                self.current_language = language
                out.write(colours.color_language)
                out.write(f"[{language}]")
                out.write(colours.color_reset)

    def _format_word_with_probability(
        self,
        *,
        out: TextIO,
        word: str,
        probability: float,
        utterance_probability: float,
        colours: ColorScheme,
    ):
        # pylint: disable=too-many-boolean-expressions
        if (
            self.probability_style == ProbabilityStyle.word
            and probability < 0.90 / 2
            or self.probability_style == ProbabilityStyle.word_75
            and probability < 0.75 / 2
            or self.probability_style == ProbabilityStyle.word_50
            and probability < 0.50 / 2
            or self.probability_style == ProbabilityStyle.word_25
            and probability < 0.25 / 2
        ):
            out.write(colours.color_text_verylowconfidence)
            out.write(word)
            out.write(colours.color_reset)
        elif (
            self.probability_style == ProbabilityStyle.word
            and probability < 0.90
            or self.probability_style == ProbabilityStyle.word_75
            and probability < 0.75
            or self.probability_style == ProbabilityStyle.word_50
            and probability < 0.50
            or self.probability_style == ProbabilityStyle.word_25
            and probability < 0.25
        ):
            out.write(colours.color_text_lowconfidence)
            out.write(word)
            out.write(colours.color_reset)
        elif (
            self.probability_style == ProbabilityStyle.line
            and utterance_probability < 0.90 / 2
            or self.probability_style == ProbabilityStyle.line_75
            and utterance_probability < 0.75 / 2
            or self.probability_style == ProbabilityStyle.line_50
            and utterance_probability < 0.50 / 2
            or self.probability_style == ProbabilityStyle.line_25
            and utterance_probability < 0.25 / 2
        ):
            out.write(colours.color_text_verylowconfidence)
            out.write(word)
            out.write(colours.color_reset)
        elif (
            self.probability_style == ProbabilityStyle.line
            and utterance_probability < 0.90
            or self.probability_style == ProbabilityStyle.line_75
            and utterance_probability < 0.75
            or self.probability_style == ProbabilityStyle.line_50
            and utterance_probability < 0.50
            or self.probability_style == ProbabilityStyle.line_25
            and utterance_probability < 0.25
        ):
            out.write(colours.color_text_lowconfidence)
            out.write(word)
            out.write(colours.color_reset)
        else:
            out.write(colours.color_text)
            out.write(word)
            out.write(colours.color_reset)

    def format_utterance(self, utterance: Utterance, out: TextIO, colours: ColorScheme):
        self._format_timestamp(out=out, start_ts=utterance.start_ts, end_ts=utterance.end_ts, colours=colours)
        self._format_speaker(out=out, speaker=utterance.speaker or UNKNOWN_SPEAKER, colours=colours)

        percentile_25 = float(np.percentile([w.probability for w in utterance.words], 25))

        # pylint: disable=superfluous-parens
        for i, w in enumerate(utterance.words):
            self._format_language(out=out, language=w.lang, first_word=(i == 0), colours=colours)
            self._format_word_with_probability(out=out, word=w.word, probability=w.probability, utterance_probability=percentile_25, colours=colours)
        out.write("\n")


class TextIOTranscriptWriter(TranscriptWriter):
    out: TextIO

    def __init__(
        self,
        *,
        config: TranscriptWriterConfig,
        acknowledged_colours=COLORSCHEME_ACKNOWLEDGED,
        unacknowledged_colours=COLORSCHEME_UNACKNOWLEDGED,
        unconfirmed_colors=COLORSCHEME_UNCONFIRMED,
        print_unacknowledged: bool = False,
    ):
        super().__init__(config)
        self.formatter: TranscriptFormatter = TranscriptFormatter(
            language_style=config.language_style,
            probability_style=config.probability_style,
            speaker_style=config.speaker_style,
            timestamp_style=config.timestamp_style,
        )
        self.acknowledged_colours = acknowledged_colours
        self.unacknowledged_colours = unacknowledged_colours
        self.unconfirmed_colors = unconfirmed_colors
        self.print_unacknowledged = print_unacknowledged

    def _set_textio(self, out: TextIO):
        self.out = out

    def open(self, path_no_ext: str):
        pass

    def close(self):
        pass

    def write(
        self,
        utterance: Utterance,
        unacknowledged_utterance: Optional[List[Utterance]] = None,
        unconfirmed_words: Optional[List[Word]] = None,
    ):
        self.formatter.format_utterance(utterance=utterance, out=self.out, colours=self.acknowledged_colours)
        if self.print_unacknowledged:
            if unacknowledged_utterance:
                for unack in unacknowledged_utterance:
                    self.formatter.format_utterance(utterance=unack, out=self.out, colours=self.unconfirmed_colors)
            if unconfirmed_words and len(unconfirmed_words) > 0:
                self.formatter.format_utterance(
                    utterance=Utterance.from_words(utterance_id="", words=unconfirmed_words),
                    out=self.out,
                    colours=self.unconfirmed_colors,
                )
        self.out.flush()


class TextTranscriptWriter(TextIOTranscriptWriter):
    def __init__(self, config: TranscriptWriterConfig):
        super().__init__(
            config=config,
            acknowledged_colours=COLORSCHEME_NONE,
            unacknowledged_colours=COLORSCHEME_NONE,
            unconfirmed_colors=COLORSCHEME_NONE,
        )

    def open(self, path_no_ext: str):
        # pylint: disable=consider-using-with
        self._set_textio(open(f"{path_no_ext}.txt", "w", encoding="utf-8"))

    def close(self):
        self.out.close()



import logging
import math
import re
from abc import ABC, abstractmethod
from time import perf_counter
from typing import List

from verbatim.transcript.words import Word
from verbatim_audio.audio import samples_to_seconds

# Configure logger
LOG = logging.getLogger(__name__)


class SentenceTokenizer(ABC):
    @abstractmethod
    def split(self, words: List[Word]) -> List[str]:
        pass


class FastSentenceTokenizer(SentenceTokenizer):
    def split(self, words: List[Word]) -> List[str]:
        text = "".join(w.word for w in words)
        # List of punctuation marks to split on
        split_punctuations = r".。;!！?？"
        punctuations = "\"'.。,;，!！?？:：”)]}、\"'“¿([{-"
        # Create a regex pattern to match any punctuation character
        regex_punctuation = f"([{re.escape(split_punctuations)}])"
        regex_not_punctuation = f"([^{re.escape(punctuations)}])"
        # Split the sentence, keeping the punctuation in the result
        parts = re.split(regex_punctuation, text)
        # Merge punctuation with the preceding part, preserving spaces
        result = []
        text_words = ""
        for part in parts:
            text_words += part
            if text_words and re.match(regex_not_punctuation, text_words):
                result.append(text_words)
                text_words = ""
            elif len(text_words) > 0:
                result[-1] += text_words
                text_words = ""

        if text_words:
            if re.match(regex_not_punctuation, text_words):
                result.append(text_words)
            elif len(result) > 0:
                result[-1] += text_words
            else:
                result.append(text_words)

        return result


class SaTSentenceTokenizer(SentenceTokenizer):
    def __init__(self, device: str, model="sat-3l-sm"):
        LOG.info("Lazy-loading SaT Sentence Tokenizer (model=%s)", model)
        start = perf_counter()
        from wtpsplit import SaT  # pylint: disable=import-outside-toplevel

        self.sat_sm = SaT(model)
        self.sat_sm.half().to(device)
        LOG.info("SaT Sentence Tokenizer ready in %.2fs", perf_counter() - start)

    def split(self, words: List[Word]) -> List[str]:
        text = "".join(w.word for w in words)
        return self.sat_sm.split(text)  # pyright: ignore[reportReturnType]


class SilenceSentenceTokenizer(SentenceTokenizer):
    def __init__(self):
        # parameters can be tuned as needed
        self.target_duration = 10.0  # target sentence length in seconds
        self.min_duration = 2.0  # below this, the penalty is huge
        self.max_duration = 25.0  # above this, also huge penalty
        # Weight for how much a silence gap improves the candidate break.
        self.gap_weight = 1.0
        self.distance_weight = 1.0
        self.silence_weight = 2.0
        self.word_weight = 1.0

    def split(self, words: List[Word]) -> List[str]:
        if not words:
            return []

        def distance_energy(val: float, min_val: float, max_val: float, ideal_val: float) -> float:
            value: float = max(0, min(1, (val - min_val) / (ideal_val - min_val) if val < ideal_val else (max_val - val) / (max_val - ideal_val)))
            return 1 - value

        def silence_energy(duration: float, max_duration: float) -> float:
            value: float = max(0, min(1, duration / max_duration))
            return 1 - value

        def word_energy(left_word: str, right_word: str) -> float:
            primary_suffix_punctuations = tuple([".", "。", ";", "!", "！", "?", "？"])
            secondary_prefix_punctuations = tuple(['"', "“", "¿", "¡", "(", "[", "{", "«"])
            secondary_suffix_punctuations = tuple([",", "、"])

            value: float = 0
            if left_word.rstrip().endswith(primary_suffix_punctuations):
                value += 1
            if left_word.rstrip().endswith(secondary_suffix_punctuations):
                value += 0.25
            if right_word.lstrip().startswith(secondary_prefix_punctuations):
                value += 0.25
            return 1 - max(0, min(1, value))

        def _energy(first_word: Word, left_word: Word, right_word: Word):
            seconds_from_start = samples_to_seconds(right_word.start_ts - first_word.start_ts)
            silence_duration = samples_to_seconds(right_word.start_ts - left_word.end_ts)

            distance_cost = distance_energy(
                val=seconds_from_start, min_val=self.min_duration, max_val=self.max_duration, ideal_val=self.target_duration
            )
            silence_cost = silence_energy(duration=silence_duration, max_duration=self.max_duration)
            word_cost = word_energy(left_word=left_word.word, right_word=right_word.word)
            energy = self.distance_weight * distance_cost + self.silence_weight * silence_cost + self.word_weight * word_cost
            return energy

        sentences = []
        weights = []
        current_start = 0  # index of the first word in the current sentence

        best_break_index = None
        best_energy = math.inf

        for i in range(current_start, len(words) - 1):
            energy = _energy(first_word=words[0], left_word=words[i], right_word=words[i + 1])
            weights.append(energy)
            if energy < best_energy:
                best_energy = energy
                best_break_index = i

        LOG.debug(f"split weight:\n{''.join(w.word for w in words)}\n{weights}")
        # If no candidate was found (e.g. because the duration grows too large)
        # then take all remaining words.
        if best_break_index is None:
            best_break_index = len(words) - 1

        # Build the sentence from current_start to best_break_index (inclusive)
        sentence = "".join(word.word for word in words[current_start : best_break_index + 1])
        if sentence:
            sentences.append(sentence)

        sentence = "".join(word.word for word in words[best_break_index + 1 :])
        if sentence:
            sentences.append(sentence)

        # Move to the next segment
        current_start = best_break_index + 1

        return sentences


class BoundedSentenceTokenizer(SentenceTokenizer):
    def __init__(
        self, other_tokenizer: SentenceTokenizer, bounding_tokenizer: SentenceTokenizer = SilenceSentenceTokenizer(), max_duration: float = 25
    ):
        self.other_tokenizer = other_tokenizer
        self.bounding_tokenizer = bounding_tokenizer
        self.max_duration = max_duration

    def split(self, words: List[Word]) -> List[str]:
        tok = self.other_tokenizer.split(words=words)
        if len(tok) != 1:
            return tok

        duration: float = samples_to_seconds(words[-1].end_ts - words[0].start_ts)
        if duration <= self.max_duration:
            return tok

        return self.bounding_tokenizer.split(words=words)



# PS08_verbatim/verbatim/voices/transcribe/whispercpp.py

import logging
import multiprocessing
import sys
from typing import List, Tuple, Union

from numpy.typing import NDArray
from pywhispercpp.model import Model, Segment

from verbatim_audio.audio import samples_to_seconds, seconds_to_samples

from ...transcript.words import Word
from .transcribe import Transcriber

LOG = logging.getLogger(__name__)


class WhisperCppTranscriber(Transcriber):
    def __init__(
        self,
        *,
        model_size_or_path: str,
        device: str,
        whisper_beam_size: int = 3,
        whisper_best_of: int = 3,
        whisper_patience: float = 1.0,
        whisper_temperatures: Union[None, List[float]] = None,
    ):
        if whisper_temperatures is None:
            whisper_temperatures = [0.0, 0.2, 0.4, 0.6, 0.8, 1.0]

        if device != "cpu":
            # WhisperCPP doesn't support GPU
            LOG.warning("WhisperCPP only supports CPU inference, ignoring GPU device")

        self.whisper_model = Model(model=model_size_or_path, redirect_whispercpp_logs_to=sys.stderr)

        self.whisper_beam_size = whisper_beam_size
        self.whisper_best_of = whisper_best_of
        self.whisper_patience = whisper_patience
        self.whisper_temperatures = whisper_temperatures

    def guess_language(self, audio: NDArray, lang: List[str]) -> Tuple[str, float]:
        # https://github.com/gaspardpetit/verbatim/issues/78
        # For now return first language
        return lang[0], 1.0

    def transcribe(
        self,
        *,
        audio: NDArray,
        lang: str,
        prompt: str,
        prefix: str,
        window_ts: int,
        audio_ts: int,
        whisper_beam_size: int = 3,
        whisper_best_of: int = 3,
        whisper_patience: float = 1.0,
        whisper_temperatures: Union[None, List[float]] = None,
    ) -> List[Word]:
        LOG.info(f"Transcription Prefix: {prefix}")

        transcript_words: List[Word] = []

        def on_segment(w: Segment):
            if w.text.strip() == "":
                return

            # WhisperCPP segments are in centiseconds (1/100 second)
            start_ts = seconds_to_samples(w.t0 / 100) + window_ts  # Convert to samples
            end_ts = seconds_to_samples(w.t1 / 100) + window_ts

            if end_ts > audio_ts:
                return

            word = Word.from_whisper_cpp_1w_segment(segment=w, lang=lang, ts_offset=window_ts)

            LOG.debug(f"[{start_ts} ({samples_to_seconds(start_ts)})]: {word.word}")
            transcript_words.append(word)

        cpu_count = max(1, multiprocessing.cpu_count() // 2)
        # Run inference
        self.whisper_model.transcribe(
            audio,
            max_len=1,
            split_on_word=True,
            token_timestamps=True,
            initial_prompt=prompt,
            language=lang,
            new_segment_callback=on_segment,
            n_threads=cpu_count,
            no_context=True,
            # When using these, whisper.cpp complaints about too many encoders - disabled for now
            # greedy={"best_of": whisper_best_of},
            # beam_search={'beam_size': whisper_beam_size, 'patience': whisper_patience},
        )

        LOG.debug(transcript_words)
        return transcript_words



from typing import Optional

import numpy as np
from numpy.typing import NDArray

from ..audio import format_audio
from .audiosource import AudioSource, AudioStream

Annotation = object  # pylint: disable=invalid-name


def _to_float32(audio: NDArray) -> NDArray:
    if audio.dtype == np.float32:
        return audio
    if audio.dtype == np.int8:
        return audio.astype(np.float32) / 128.0
    if audio.dtype == np.int16:
        return audio.astype(np.float32) / 32768.0
    if audio.dtype == np.int32:
        return audio.astype(np.float32) / 2147483648.0
    return audio.astype(np.float32)


class MemoryAudioStream(AudioStream):
    source: "MemoryAudioSource"

    def __init__(self, source: "MemoryAudioSource", diarization: Optional[Annotation]):
        super().__init__(start_offset=source.start_sample, diarization=diarization)
        self.source = source
        self._cursor = source.start_sample

    def next_chunk(self, chunk_length=1) -> NDArray:
        if not self.has_more():
            return np.array([], dtype=np.float32)

        chunk_samples = int(chunk_length * self.source.sample_rate)
        end_sample = min(self._cursor + chunk_samples, self.source.end_sample or self.source.total_samples)
        raw = self.source.samples[self._cursor : end_sample]
        self._cursor = end_sample

        if raw.size == 0:
            return np.array([], dtype=np.float32)

        if self.source.preserve_channels:
            return _to_float32(raw)

        # Mix down and resample to target rate via format_audio
        return format_audio(_to_float32(raw), from_sampling_rate=self.source.sample_rate)

    def has_more(self) -> bool:
        end = self.source.end_sample or self.source.total_samples
        return self._cursor < end

    def close(self):
        pass

    def get_nchannels(self) -> int:
        if self.source.samples.ndim == 1:
            return 1
        return self.source.samples.shape[1]

    def get_rate(self) -> int:
        return self.source.sample_rate


class MemoryAudioSource(AudioSource):
    samples: NDArray
    sample_rate: int
    diarization: Optional[Annotation]
    preserve_channels: bool
    start_sample: int
    end_sample: Optional[int]
    total_samples: int

    def __init__(
        self,
        *,
        samples: NDArray,
        sample_rate: int,
        diarization: Optional[Annotation] = None,
        preserve_channels: bool = False,
        channels_first: bool = False,
        start_sample: int = 0,
        end_sample: Optional[int] = None,
        source_name: str = "<memory>",
    ):
        super().__init__(source_name=source_name)
        if channels_first and samples.ndim == 2:
            samples = samples.T
        self.samples = samples
        self.sample_rate = sample_rate
        self.diarization = diarization
        self.preserve_channels = preserve_channels
        self.start_sample = max(0, start_sample)
        self.end_sample = end_sample
        self.total_samples = samples.shape[0] if samples.ndim >= 1 else 0

    def open(self) -> MemoryAudioStream:
        return MemoryAudioStream(source=self, diarization=self.diarization)



import io
import logging
import os
from typing import Any, Dict, List, Optional, Tuple, cast

# pylint: disable=import-outside-toplevel,broad-exception-caught
import numpy as np
import torch
from pyannote.audio import Pipeline
from pyannote.audio.core.task import Problem, Resolution, Specifications
from pyannote.audio.pipelines.utils.hook import ProgressHook
from torch.serialization import add_safe_globals

from verbatim.cache import ArtifactCache
from verbatim_audio.audio import wav_to_int16
from verbatim_audio.sources.audiosource import AudioSource
from verbatim_audio.sources.fileaudiosource import FileAudioSource
from verbatim_diarization.separate.base import SeparationStrategy
from verbatim_diarization.utils import sanitize_uri_component
from verbatim_files.rttm import Annotation as RTTMAnnotation
from verbatim_files.rttm import Segment, dumps_rttm
from verbatim_files.vttm import AudioRef, dumps_vttm

from .constants import PYANNOTE_SEPARATION_MODEL_ID
from .ffmpeg_loader import ensure_torchcodec_audio_decoder

# Configure logger
LOG = logging.getLogger(__name__)


def _build_rttm_annotation(diarization, label_to_ref: Dict[str, AudioRef], default_uri: str) -> RTTMAnnotation:
    segments = []
    for segment, _track, label in diarization.itertracks(yield_label=True):
        seg_label = str(label)
        audio_ref = label_to_ref.get(seg_label)
        file_id = audio_ref.id if audio_ref else default_uri
        segments.append(Segment(start=segment.start, end=segment.end, speaker=seg_label, file_id=file_id))
    return RTTMAnnotation(segments=segments, file_id=default_uri)


class PyannoteSpeakerSeparation(SeparationStrategy):
    def __init__(
        self,
        *,
        cache: ArtifactCache,
        device: str,
        huggingface_token: str,
        separation_model=PYANNOTE_SEPARATION_MODEL_ID,
        diarization_strategy: str = "pyannote",
        **kwargs,
    ):
        super().__init__(cache=cache)
        del kwargs  # unused
        LOG.info("Initializing Separation Pipeline.")
        self.diarization_strategy = diarization_strategy
        self.device = device
        self.huggingface_token = huggingface_token
        # Allow safe loading of pyannote checkpoints under torch>=2.6
        torch_version_mod: Any = getattr(torch, "torch_version", None)
        safe_types = [Specifications, Problem, Resolution]
        if torch_version_mod is not None:
            safe_types.append(torch_version_mod.TorchVersion)  # type: ignore[attr-defined]
        add_safe_globals(safe_types)  # pyright: ignore[reportArgumentType]

        self.pipeline = Pipeline.from_pretrained(separation_model, token=self.huggingface_token)
        hyper_parameters = {
            "segmentation": {"min_duration_off": 0.0, "threshold": 0.82},
            "clustering": {
                "method": "centroid",
                "min_cluster_size": 15,
                "threshold": 0.68,
            },
            "separation": {
                "leakage_removal": True,
                "asr_collar": 0.32,
            },
        }

        if self.pipeline is None:
            raise RuntimeError("Pyannote separation pipeline failed to initialize")

        self.pipeline.instantiate(hyper_parameters)
        self.pipeline.to(torch.device(device))

    def __enter__(self) -> "PyannoteSpeakerSeparation":
        return self

    def __exit__(self, ex_type, ex_value, ex_traceback):
        del self.pipeline
        return False

    def _prepare_pipeline_input(self, file_path: str) -> Tuple[Any, Optional[str]]:
        temp_path: Optional[str] = None
        try:
            import soundfile as sf  # lazy import

            buffer = self.cache.bytes_io(file_path)
            audio, sample_rate = sf.read(buffer)

            if isinstance(audio, np.ndarray) and audio.ndim > 1 and audio.shape[1] > 1:
                mono = np.mean(audio, axis=1)
            else:
                mono = audio

            waveform = torch.from_numpy(np.asarray(mono, dtype=np.float32))
            if waveform.ndim == 1:
                waveform = waveform.unsqueeze(0)
            return {"waveform": waveform, "sample_rate": int(sample_rate)}, None
        except Exception:
            temp_path = None
        return file_path, temp_path

    def _separate_to_audio_refs(
        self,
        *,
        file_path: str,
        out_speaker_wav_prefix: str,
        nb_speakers: Optional[int],
    ) -> Tuple[Any, List[Tuple[str, AudioRef]]]:
        ensure_torchcodec_audio_decoder("pyannote separation")
        file_for_pipeline, temp_path = self._prepare_pipeline_input(file_path)
        try:
            with ProgressHook() as hook:
                if self.pipeline is None:
                    raise RuntimeError("Pyannote separation pipeline is not initialized")
                diarization_output, sources = self.pipeline(file_for_pipeline, hook=hook, num_speakers=nb_speakers)
        finally:
            if temp_path and temp_path != file_path and os.path.exists(temp_path):
                try:
                    os.unlink(temp_path)
                except Exception:  # pragma: no cover
                    LOG.debug("Failed to clean up temporary separation file %s", temp_path)

        diarization = diarization_output.speaker_diarization if hasattr(diarization_output, "speaker_diarization") else diarization_output

        sources_data = np.asarray(sources.data)
        shape = cast(tuple[int, int], tuple(sources_data.shape[:2]))

        audio_refs_meta: List[Tuple[str, AudioRef]] = []
        for s, speaker in enumerate(diarization.labels()):
            if s >= shape[1]:
                LOG.debug("Skipping speaker index %s as it is out of bounds for separated sources", s)
                continue

            speaker_data = sources_data[:, s]
            if speaker_data.dtype != np.int16:
                speaker_data = wav_to_int16(speaker_data)

            file_name = f"{out_speaker_wav_prefix}-{speaker}.wav" if out_speaker_wav_prefix else f"{speaker}.wav"
            import soundfile as sf  # lazy import

            buffer = io.BytesIO()
            sf.write(buffer, speaker_data, 16000, format="WAV")
            self.cache.set_bytes(file_name, buffer.getvalue())

            sanitized_id = sanitize_uri_component(os.path.splitext(os.path.basename(file_name))[0])
            audio_ref = AudioRef(id=sanitized_id, path=file_name, channels=None)
            audio_refs_meta.append((str(speaker), audio_ref))

        if not audio_refs_meta:
            uri = sanitize_uri_component(os.path.splitext(os.path.basename(file_path))[0])
            audio_refs_meta.append((uri, AudioRef(id=uri, path=file_path, channels=None)))

        return diarization, audio_refs_meta

    # pylint: disable=unused-argument
    def separate_speakers(
        self,
        *,
        file_path: str,
        out_rttm_file: Optional[str] = None,
        out_vttm_file: Optional[str] = None,
        out_speaker_wav_prefix="",
        nb_speakers: Optional[int] = None,
        start_sample: int = 0,
        end_sample: Optional[int] = None,
    ) -> List[AudioSource]:
        """
        Separate speakers in an audio file.

        Args:
            file_path: Path to input audio file
            out_rttm_file: Optional legacy RTTM output path
            out_speaker_wav_prefix: Prefix for output WAV files
            nb_speakers: Optional number of speakers

        Returns:
            List of FileAudioSource entries, one per separated speaker.
        """
        diarization, audio_refs_meta = self._separate_to_audio_refs(
            file_path=file_path,
            out_speaker_wav_prefix=out_speaker_wav_prefix,
            nb_speakers=nb_speakers,
        )

        uri = sanitize_uri_component(os.path.splitext(os.path.basename(file_path))[0])
        label_to_ref = dict(audio_refs_meta)
        diarization_annotation = _build_rttm_annotation(diarization, label_to_ref, uri)

        separated_sources: List[AudioSource] = []
        for _label, audio_ref in audio_refs_meta:
            separated_sources.append(
                FileAudioSource(
                    file=audio_ref.path,
                    cache=self.cache,
                    start_sample=start_sample,
                    end_sample=end_sample,
                    diarization=diarization_annotation,
                    file_id=audio_ref.id,
                )
            )

        if out_rttm_file:
            self.cache.set_text(out_rttm_file, dumps_rttm(diarization_annotation))

        if out_vttm_file:
            audio_refs = [ref for _label, ref in audio_refs_meta]
            self.cache.set_text(out_vttm_file, dumps_vttm(audio=audio_refs, annotation=diarization_annotation))

        return separated_sources
